import snowflake.connector
import os
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import dsa
from cryptography.hazmat.primitives import serialization
from logs.manager import LoggingManager
import getpass
import traceback


class PochiUtil:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(PochiUtil, cls).__new__(cls)
            cls._instance.init_data()
        return cls._instance

    def init_data(self):
        self.__connection = None

    def get_connection(self, options):
        has_errors = False
        try:
            # 1. self.__connection already exists; return (no errors)
            if self.__connection is not None:
                return has_errors
            
            # 2. default_connection is not defined in the config/project.toml. return error.
            if options.default_connection is None:
                LoggingManager.display_message(
                    "connection_file_missing",
                )
                has_errors = True
                return has_errors
            
            # 3. default_connection is defined in the config/project.toml, but the name is not defined in the
            #    connection config file. return error
            if (not options.default_connection.is_defined):
                LoggingManager.display_message(
                    "invalid_connection_name_issue",
                    [options.project_config.default_connection,
                    options.default_connection.connection_file_path]
                )
                has_errors = True
                return has_errors

            # 4. if accountname is missing, or username is missing, return error.
            if ("account" not in options.default_connection or
                 "user" not in options.default_connection):
                LoggingManager.display_message(
                            "missing_parameters_connection_issue",
                        )

                has_errors = True
                return has_errors
            
            # authentication sequence:
            # if private_key_path exists, then use private_key_path.
            # otherwise, if password exists, then use the password, else ask for password
            if ("private_key_path" in options.default_connection):
                # this connection is using a private_key_path
                if (os.getenv ("SNOWSQL_PRIVATE_KEY_PASSPHRASE") is None):
                    LoggingManager.display_message(
                            "missing_private_key_passphrase_connection_issue",
                        )
                    has_errors = True
                    return has_errors
                
                with open(options.default_connection.private_key_path, "rb") as key:
                    p_key = serialization.load_pem_private_key(
                        key.read(),
                        password=os.getenv("SNOWSQL_PRIVATE_KEY_PASSPHRASE").encode(),
                        backend=default_backend(),
                    )

                pkb = p_key.private_bytes(
                    encoding=serialization.Encoding.DER,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.NoEncryption(),
                )

                self.__connection = snowflake.connector.connect(
                    user=options.default_connection.user,
                    account=options.default_connection.account,
                    private_key=pkb,
                    role=getattr(options.default_connection, "role", None),
                    warehouse=getattr(options.default_connection, "warehouse", None)
                )
            else:
                # this connection has user and password defined in the config. Use them directly.
                connection_password = None
                if ("password" in options.default_connection):
                    connection_password = options.default_connection.password
                else:
                    connection_password = getpass.getpass(prompt="Enter Password for Connection " + options.project_config.default_connection + ": ")

                self.__connection = snowflake.connector.connect(
                    user=options.default_connection.user,
                    password=connection_password,
                    account=options.default_connection.account,
                    role=getattr(options.default_connection, "role", None),
                    warehouse=getattr(options.default_connection, "warehouse", None)
                )

        except snowflake.connector.Error as e:
            LoggingManager.display_message(
                "connection_issues",
                [
                    options.default_connection.account,
                    options.project_config.default_connection,
                ],
            )
            LoggingManager.display_single_message(
                e
            )
            has_errors = True
        except Exception as e:
            LoggingManager.display_single_message(
                f"Unexpected {type(e)=}: {e=}"
            )
            LoggingManager.display_single_message(traceback.format_exc())
            has_errors = True
        finally:
            return has_errors

    def execute_sql(self, sql_statement, with_output=False):
        has_errors = False
        try:
            cur = self.__connection.cursor().execute(sql_statement)
            if self.__connection.get_query_status(cur.sfqid).name != "SUCCESS":
                has_errors = True
        except Exception as e:
            LoggingManager.display_message(
                "script_issues",
                [
                    sql_statement,
                    e,
                ],
            )
            LoggingManager.display_single_message(traceback.format_exc())
            has_errors = True
        finally:
            if with_output:
                return has_errors, cur.fetchall()
            return has_errors

    def execute_sql_from_file(self, file_path, has_errors=False, query_logging=False):
        try:
            if os.path.exists(file_path) and not has_errors:
                with open(file_path, "r") as sql_file:
                    for cur in self.__connection.execute_stream(
                        sql_file, remove_comments=True
                    ):
                        if query_logging:
                            col_width = 39 if len(cur.description) > 1 else 121
                            LoggingManager.display_single_message(
                                "[SQL] +-"
                                + "+-".join("-" * col_width for col in cur.description)
                                + "+"
                            )
                            LoggingManager.display_single_message(
                                # "[SQL] | " + "| ".join(str(col.name)[:col_width].ljust(col_width) for col in cur.description) + "|"
                                "[SQL] | "
                                + "| ".join(
                                    str(col.name)[:col_width].ljust(col_width)
                                    for col in cur.description
                                )
                                + "|"
                            )
                            LoggingManager.display_single_message(
                                "[SQL] +-"
                                + "+-".join("-" * col_width for col in cur.description)
                                + "+"
                            )
                            for ret in cur:
                                LoggingManager.display_single_message(
                                    # "[SQL] | " + "| ".join(str(col)[:col_width].ljust(col_width) for col in ret) + "|"
                                    "[SQL] | "
                                    + "| ".join(
                                        str(col).ljust(col_width) for col in ret
                                    )
                                    + "|"
                                )
                            LoggingManager.display_single_message(
                                "[SQL] +-"
                                + "+-".join("-" * col_width for col in cur.description)
                                + "+"
                            )
                            LoggingManager.display_single_message("[SQL] ")
                        if (
                            self.__connection.get_query_status(cur.sfqid).name
                            != "SUCCESS"
                        ):
                            has_errors = True
        except snowflake.connector.Error as e:
            LoggingManager.display_message(
                "script_issues",
                [
                    file_path,
                    e,
                ],
            )
            has_errors = True
        except Exception as e:
            LoggingManager.display_message(
                "script_issues",
                [
                    file_path,
                    e,
                ],
            )
            LoggingManager.display_single_message(traceback.format_exc())
            has_errors = True
        finally:
            return has_errors
